/*
 * This file is part of MPlayer.
 *
 * MPlayer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MPlayer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with MPlayer; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "config.h"

// Initial draft of my new cache system...
// Note it runs in 2 processes (using fork()), but doesn't require locking!!
// TODO: seeking, data consistency checking

#if CONFIG_STREAM_CACHE

#define DEBUG(x) ;

#define READ_SLEEP_TIME 10
// These defines are used to reduce the cost of many successive
// seeks (e.g. when a file has no index) by spinning quickly at first.
#define INITIAL_FILL_USLEEP_TIME 1000
#define INITIAL_FILL_USLEEP_COUNT 10
#define FILL_USLEEP_TIME 50000
#define PREFILL_SLEEP_TIME 200
#define CONTROL_SLEEP_TIME 0

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include "libavutil/avutil.h"
#include "osdep/shmem.h"
#include "osdep/timer.h"
#if defined(__MINGW32__)
#include <windows.h>
static void ThreadProc( void *s );
#elif defined(__OS2__)
#define INCL_DOS
#include <os2.h>
static void ThreadProc( void *s );
#elif defined(PTHREAD_CACHE)
#include <pthread.h>
static void *ThreadProc(void *s);
#elif defined(__MORPHOS__)
#else
#include <sys/wait.h>
#define FORKED_CACHE 1
#endif
#ifndef FORKED_CACHE
#define FORKED_CACHE 0
#endif

#include "mp_msg.h"
#include "help_mp.h"

#include "stream.h"
#include "cache2.h"

#ifdef __MORPHOS__

#include <errno.h>
#include <clib/debug_protos.h>
#include <exec/ports.h>
#include <exec/tasks.h>
#include <exec/types.h>
#include <dos/dostags.h>
//#include <amitcp/socketbasetags.h>

#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/socket.h>

#include "input/input.h"
#include "morphos_stuff.h"

#if MPLAYER
#ifdef CONFIG_GUI
#include "gui/interface.h"
#include "gui/morphos/gui.h"
extern int use_gui;
extern char * stream_dump_name;
static FILE * dump_f = NULL;
#endif
#endif

extern struct Library * SocketBase;
extern struct Task * maintask;

struct Library * oldSocketBase;
struct Process * cachetask;
static LONG fdkey;

#define CURRENT_TASK (FindTask(NULL) == maintask) ? "[MAIN TASK]" : "[CACHE TASK]"

#endif

int stream_fill_buffer(stream_t *s);
int stream_seek_long(stream_t *s,quad_t pos);

typedef struct {
  // constats:
  unsigned char *buffer;      // base pointer of the allocated buffer memory
  int buffer_size; // size of the allocated buffer memory
  int sector_size; // size of a single sector (2048/2324)
  int back_size;   // we should keep back_size amount of old bytes for backward seek
  int fill_limit;  // we should fill buffer only if space>=fill_limit
  int seek_limit;  // keep filling cache if distance is less that seek limit
  // filler's pointers:
  int eof;
  quad_t min_filepos; // buffer contain only a part of the file, from min-max pos
  quad_t max_filepos;
  quad_t offset;      // filepos <-> bufferpos  offset value (filepos of the buffer's first byte)
  // reader's pointers:
  quad_t read_filepos;
  // commands/locking:
//  int seek_lock;   // 1 if we will seek/reset buffer, 2 if we are ready for cmd
//  int fifo_flag;  // 1 if we should use FIFO to notice cache about buffer reads.
  // callback
  stream_t* stream;
  volatile int control;
  volatile unsigned control_uint_arg;
  volatile double control_double_arg;
  volatile int control_res;
  volatile quad_t control_new_pos;
  volatile double stream_time_length;
  volatile double stream_current_time; //__MORPHOS__
} cache_vars_t;

static int min_fill=0;

int cache_fill_status=0;

#ifdef __MORPHOS__

void force_exit(void)
{
	/*
	if(cachetask)
	{
		Signal((struct Task *) cachetask, SIGBREAKF_CTRL_E);
	}
	*/
}

#endif

static void cache_wakeup(stream_t *s)
{
#if FORKED_CACHE
  // signal process to wake up immediately
  kill(s->cache_pid, SIGUSR1);
#endif
}

static int cache_read(cache_vars_t *s, unsigned char *buf, int size)
{
  int total=0;
  int sleep_count = 0;
  int last_max = s->max_filepos;

  DEBUG(kprintf("%s cache_read\n", CURRENT_TASK));

  while(size>0){
	int pos,newb,len;

  //printf("CACHE2_READ: 0x%X <= 0x%X <= 0x%X  \n",s->min_filepos,s->read_filepos,s->max_filepos);

	// In case cache stalls for some reason, let's accept to break it
	if (SetSignal(0L, SIGBREAKF_CTRL_C|SIGBREAKF_CTRL_E) & (SIGBREAKF_CTRL_C|SIGBREAKF_CTRL_E))
	{
		//s->eof = 1;
		//break;
		return 0;
	}
	
	if(s->read_filepos>=s->max_filepos || s->read_filepos<s->min_filepos){
	// eof?
	if(s->eof) break;
	if (s->max_filepos == last_max) {
	    if (sleep_count++ == 10)
	        mp_msg(MSGT_CACHE, MSGL_WARN, "Cache not filling!\n");
	} else {
	    last_max = s->max_filepos;
	    sleep_count = 0;
	}
	// waiting for buffer fill...
	if (stream_check_interrupt(READ_SLEEP_TIME)) {
	    s->eof = 1;
	    break;
	}
	continue; // try again...
	}
    sleep_count = 0;

	newb=s->max_filepos-s->read_filepos; // new bytes in the buffer
	if(newb<min_fill) min_fill=newb; // statistics...

//    printf("*** newb: %d bytes ***\n",newb);

	pos=s->read_filepos - s->offset;
	if(pos<0) pos+=s->buffer_size; else
	if(pos>=s->buffer_size) pos-=s->buffer_size;

    if(newb>s->buffer_size-pos) newb=s->buffer_size-pos; // handle wrap...
    if(newb>size) newb=size;

    // check:
    if(s->read_filepos<s->min_filepos) mp_msg(MSGT_CACHE,MSGL_ERR,"Ehh. s->read_filepos<s->min_filepos !!! Report bug...\n");

    // len=write(mem,newb)
    //printf("Buffer read: %d bytes\n",newb);
    memcpy(buf,&s->buffer[pos],newb);
    buf+=newb;
    len=newb;
    // ...

    s->read_filepos+=len;
    size-=len;
    total+=len;
  }

  DEBUG(kprintf("%s cache_read total = %d\n", CURRENT_TASK, total));

  cache_fill_status=(s->max_filepos-s->read_filepos)/(s->buffer_size / 100);
  return total;
}

static int cache_fill(cache_vars_t *s)
{
  int back,back2,newb,space,len,pos;
  quad_t read=s->read_filepos;
  int read_chunk;

  DEBUG(kprintf("%s cache_fill read = %lld s->min_filepos = %lld s->max_filepos = %lld\n", CURRENT_TASK, read, s->min_filepos, s->max_filepos));

  if(read<s->min_filepos || read>s->max_filepos){
      // seek...
      mp_msg(MSGT_CACHE,MSGL_DBG2,"Out of boundaries... seeking to 0x%"PRIX64"  \n",(int64_t)read);
      // drop cache contents only if seeking backward or too much fwd.
      // This is also done for on-disk files, since it loses the backseek cache.
      // That in turn can cause major bandwidth increase and performance
      // issues with e.g. mov or badly interleaved files
      if(read<s->min_filepos || read>=s->max_filepos+s->seek_limit)
      {
        s->offset= // FIXME!?
        s->min_filepos=s->max_filepos=read; // drop cache content :(
        if(s->stream->eof) stream_reset(s->stream);
        stream_seek_internal(s->stream,read);
        mp_msg(MSGT_CACHE,MSGL_DBG2,"Seek done. new pos: 0x%"PRIX64"  \n",(int64_t)stream_tell(s->stream));
      }
  }

  // calc number of back-bytes:
  back=read - s->min_filepos;
  if(back<0) back=0; // strange...
  if(back>s->back_size) back=s->back_size;

  // calc number of new bytes:
  newb=s->max_filepos - read;
  if(newb<0) newb=0; // strange...

  // calc free buffer space:
  space=s->buffer_size - (newb+back);

  // calc bufferpos:
  pos=s->max_filepos - s->offset;
  if(pos>=s->buffer_size) pos-=s->buffer_size; // wrap-around

  if(space<s->fill_limit){
//    printf("Buffer is full (%d bytes free, limit: %d)\n",space,s->fill_limit);
	return 0; // no fill...
  }

//  printf("### read=0x%X  back=%d  newb=%d  space=%d  pos=%d\n",read,back,newb,space,pos);

  // reduce space if needed:
  if(space>s->buffer_size-pos) space=s->buffer_size-pos;

  // limit one-time block size
  read_chunk = s->stream->read_chunk;
  if (!read_chunk) read_chunk = 4*s->sector_size;
  space = FFMIN(space, read_chunk);

#if 1
  // back+newb+space <= buffer_size
  back2=s->buffer_size-(space+newb); // max back size
  if(s->min_filepos<(read-back2)) s->min_filepos=read-back2;
#else
  s->min_filepos=read-back; // avoid seeking-back to temp area...
#endif

  DEBUG(kprintf("%s cache_fill before read\n", CURRENT_TASK));

  len = stream_read_internal(s->stream, &s->buffer[pos], space);

#if MPLAYER
#ifdef CONFIG_GUI
  if(use_gui && mygui->dumpstream)
  {
	if(dump_f && fwrite(&s->buffer[pos],len,1,dump_f) != 1) {
		mp_msg(MSGT_CPLAYER,MSGL_STATUS,MSGTR_ErrorWritingFile,stream_dump_name);
		fclose(dump_f);
		dump_f = NULL;
	}
  }
#endif
#endif

  s->eof= !len;

  DEBUG(kprintf("%s cache_fill after read, read %d\n", CURRENT_TASK, len));
  
  s->max_filepos+=len;
  if(pos+len>=s->buffer_size){
	  // wrap...
	  s->offset+=s->buffer_size;
  }

  return len;

}

static int cache_execute_control(cache_vars_t *s) {
  static unsigned last;
  int quit = s->control == -2;
  if (quit || !s->stream->control) {
    s->stream_time_length = 0;
	s->stream_current_time = 0; // __MORPHOS__
    s->control_new_pos = 0;
    s->control_res = STREAM_UNSUPPORTED;
    s->control = -1;
    return !quit;
  }
  if (GetTimerMS() - last > 99) {
    double len;
    if (s->stream->control(s->stream, STREAM_CTRL_GET_TIME_LENGTH, &len) == STREAM_OK)
      s->stream_time_length = len;
    else
      s->stream_time_length = 0;
// __MORPHOS__
	if (s->stream->control(s->stream, STREAM_CTRL_GET_CURRENT_TIME, &len) == STREAM_OK)
	  s->stream_current_time = len;
    else
	  s->stream_current_time = 0;
// __MORPHOS__
    last = GetTimerMS();
  }
  if (s->control == -1) return 1;
  switch (s->control) {
    case STREAM_CTRL_GET_CURRENT_TIME:
    case STREAM_CTRL_SEEK_TO_TIME:
    case STREAM_CTRL_GET_ASPECT_RATIO:
      s->control_res = s->stream->control(s->stream, s->control, &s->control_double_arg);
      break;
    case STREAM_CTRL_SEEK_TO_CHAPTER:
    case STREAM_CTRL_GET_NUM_CHAPTERS:
    case STREAM_CTRL_GET_CURRENT_CHAPTER:
    case STREAM_CTRL_GET_NUM_ANGLES:
    case STREAM_CTRL_GET_ANGLE:
    case STREAM_CTRL_SET_ANGLE:
      s->control_res = s->stream->control(s->stream, s->control, &s->control_uint_arg);
      break;
    default:
      s->control_res = STREAM_UNSUPPORTED;
      break;
  }
  s->control_new_pos = s->stream->pos;
  s->control = -1;
  return 1;
}

static void *shared_alloc(int size) {
#if FORKED_CACHE
    return shmem_alloc(size);
#else
    return malloc(size);
#endif
}

static void shared_free(void *ptr, int size) {
#if FORKED_CACHE
    shmem_free(ptr, size);
#else
    free(ptr);
#endif
}

static cache_vars_t* cache_init(int size,int sector){
  int num;
  cache_vars_t* s=shared_alloc(sizeof(cache_vars_t));

  DEBUG(kprintf("%s cache_init(%d %d)\n", CURRENT_TASK, size, sector));

  if(s==NULL) return NULL;

  memset(s,0,sizeof(cache_vars_t));
  num=size/sector;
  if(num < 16){
	 num = 16;
  }//32kb min_size
  s->buffer_size=num*sector;
  s->sector_size=sector;
  s->buffer=shared_alloc(s->buffer_size);

  if(s->buffer == NULL){
    shared_free(s, sizeof(cache_vars_t));
	return NULL;
  }

  s->fill_limit=8*sector;
  s->back_size=s->buffer_size/2;

#if MPLAYER
#ifdef CONFIG_GUI
  if(use_gui && mygui->dumpstream)
  {
	  dump_f = fopen(stream_dump_name,"wb");
	
	  if(dump_f)
	  {
		mp_msg(MSGT_CPLAYER,MSGL_STATUS,"Dumping current stream to %s\n",stream_dump_name);
	  }
  }
#endif
#endif

  return s;
}

#if defined(__MORPHOS__)
extern void gdvdcss_open(void);
extern void gdvdcss_close(void);
#endif

void cache_uninit(stream_t *s) {
  cache_vars_t* c = s->cache_data;
  if(s->cache_pid) {
#if !FORKED_CACHE && !defined(__MORPHOS__)
    cache_do_control(s, -2, NULL);
#elif defined(__MORPHOS__)
    // Tell cachetask to quit
	DEBUG(kprintf("%s Signal end task\n", CURRENT_TASK));
    Signal((struct Task *) cachetask, SIGBREAKF_CTRL_E|SIGBREAKF_CTRL_C);

    // Wait for cachetask ack
	DEBUG(kprintf("%s Wait for cache task end\n", CURRENT_TASK));
    WaitPort(s->StartupMsg->Msg.mn_ReplyPort);
    while(GetMsg(s->StartupMsg->Msg.mn_ReplyPort));

    cachetask = NULL;

	DEBUG(kprintf("%s Done\n", CURRENT_TASK));

    if(s->type == STREAMTYPE_STREAM && oldSocketBase)
    {
		int fd;

		DEBUG(kprintf("%s Giving back socket to main thread\n", CURRENT_TASK));

		// Restore socketbase and descriptor
		SocketBase = oldSocketBase;
		fd = ObtainSocket(fdkey, AF_INET, SOCK_STREAM, 0);

		if(fd != -1)
		{
			s->fd = fd;
		}
		else
		{
			DEBUG(kprintf("%s Couldn't obtain socket with key %d\n", CURRENT_TASK, fdkey));
		}
    }

	DEBUG(kprintf("%s Free local stream\n", CURRENT_TASK));
    free(((cache_vars_t*)s->StartupMsg->s)->stream);

	DEBUG(kprintf("%s Delete port & message\n", CURRENT_TASK));
    DeleteMsgPort(s->StartupMsg->Msg.mn_ReplyPort);
    FreeVec(s->StartupMsg);

#if MPLAYER
#ifdef CONFIG_GUI
	  if(use_gui)
  	  {
		  if(dump_f)
		  {
		  	fclose(dump_f);
		  	dump_f = NULL;
			mp_msg(MSGT_CPLAYER,MSGL_STATUS,"Stream captured to %s\n",stream_dump_name);
	  	  }
	  }
#endif
#endif

#else /* __MORPHOS__ */
    kill(s->cache_pid,SIGKILL);
    waitpid(s->cache_pid,NULL,0);
#endif
    s->cache_pid = 0;
  }
  if(!c) return;
  shared_free(c->buffer, c->buffer_size);
  c->buffer = NULL;
  c->stream = NULL;
  shared_free(s->cache_data, sizeof(cache_vars_t));
  s->cache_data = NULL;
#ifdef __MORPHOS__
  DEBUG(kprintf("%s cache_uninit done\n", CURRENT_TASK));
#endif
}

#ifdef __MORPHOS__

void interrupt_handler(int sig)
{
	// prevent exit on a break-c, and reset errno
	errno = 0;
#if 0
	if(maintask)
	{
		Signal(maintask, SIGBREAKF_CTRL_C);
	}
#endif
}

void CacheTask(void) {

   cache_vars_t* s = NULL;
   struct CustomMsg *My_StartupMsg;
   struct Library * newSocketBase = NULL;
   int fd = -1;
   int sleep_count = 0;

   DEBUG(kprintf("%s Starting\n", CURRENT_TASK));

   signal(SIGINT, interrupt_handler);


#ifdef __amigaos4__
    My_StartupMsg = ((struct Process *)FindTask(NULL))->pr_Task.tc_UserData;
#else
   if (!(NewGetTaskAttrs(NULL, &My_StartupMsg, sizeof(struct CustomMsg *), TASKINFOTYPE_STARTUPMSG, TAG_DONE))) goto fail;
#endif



   if (!My_StartupMsg) goto fail;

   s = (cache_vars_t*) My_StartupMsg->s;

   if(s && (s->stream->type == STREAMTYPE_STREAM) && SocketBase)
   {
		// Reopen in this task context
		// It would be nicer to override socketbase name, to be really safe...
		newSocketBase = OpenLibrary("bsdsocket.library", 0);

		if(newSocketBase)
		{
			SocketBase = newSocketBase;

			//if(1)
			if(SocketBaseTags(SBTM_SETVAL(SBTC_ERRNOPTR(sizeof(errno))), (ULONG) &errno, SBTM_SETVAL(SBTC_LOGTAGPTR), (ULONG) "MPlayer", TAG_DONE) == 0)
			{
				// Get socket from main thread
				fd = ObtainSocket(fdkey, AF_INET, SOCK_STREAM, 0); /* what about sock_dgram ? */

				if(fd != -1)
				{
					// This is the new descriptor to be used. Let's hope there aren't any further calls in main thread
					s->stream->fd = fd;
				}
				else
				{
					DEBUG(kprintf("%s ObtainSocket key=%x failed\n", CURRENT_TASK, fdkey));
					goto fail;
				}
			}
			else
			{
				goto fail;
			}
		}
		else
		{
			goto fail;
		}
   }
   else if(s->stream->type == STREAMTYPE_DVD || s->stream->type == STREAMTYPE_DVDNAV)
   {
	 gdvdcss_open();
   }

//	 s->stream->fd = dup(s->stream->fd);
//	 s->stream->cache_pid = 1;

   DEBUG(kprintf("%s Starting Loop\n", CURRENT_TASK));

   while( !(SetSignal(0L, SIGBREAKF_CTRL_C|SIGBREAKF_CTRL_E) & (SIGBREAKF_CTRL_C|SIGBREAKF_CTRL_E)) )
   {
		DEBUG(kprintf("%s Loop\n", CURRENT_TASK));
		if ( !cache_fill(s) )
		{
			DEBUG(kprintf("%s Wait idle\n", CURRENT_TASK));
            if (sleep_count < INITIAL_FILL_USLEEP_COUNT) {
                sleep_count++;
				usleep(INITIAL_FILL_USLEEP_TIME);
            } else
				usleep(FILL_USLEEP_TIME); // idle
		}
		else
		{
			sleep_count = 0;
		}
        cache_execute_control((cache_vars_t*)s);
   }

   DEBUG(kprintf("%s Exiting loop\n", CURRENT_TASK));
fail:

   if(s && s->stream->type == STREAMTYPE_STREAM && newSocketBase)
   {
		// Give socket to main thread
		if(fd != -1)
		{
			DEBUG(kprintf("%s ReleaseCopyOfsocket\n", CURRENT_TASK));
			fdkey = ReleaseCopyOfSocket(s->stream->fd, time(NULL));

			if(fdkey == -1)
			{
				DEBUG(kprintf("%s ReleaseCopyOfSocket failed\n", CURRENT_TASK));
			}
		}

/*
		if (s->stream->fd >= 0)
			closesocket(s->stream->fd);
		s->stream->fd = -1;
*/
		DEBUG(kprintf("%s Closing bsdsocket\n", CURRENT_TASK));
		CloseLibrary(SocketBase);
	    SocketBase = NULL;
		newSocketBase = NULL;
   }
   else if(s && (s->stream->type == STREAMTYPE_DVD || s->stream->type == STREAMTYPE_DVDNAV))
   {
	 gdvdcss_close();
   }

   if(s)
	   s->stream->cache_pid = 0;

   if(errno != 0) errno = 0;

#ifdef __amigaos4__
    if (My_StartupMsg) ReplyMsg((struct Message *)My_StartupMsg);
#endif

   DEBUG(kprintf("%s Bye\n", CURRENT_TASK));
}

#else

   static void exit_sighandler(int x){
  // close stream
  exit(0);
}

#endif

static void dummy_sighandler(int x) {
}

/**
 * Main loop of the cache process or thread.
 */
static void cache_mainloop(cache_vars_t *s) {
    int sleep_count = 0;
#if FORKED_CACHE
    struct sigaction sa = { 0 };
    sa.sa_handler = SIG_IGN;
    sigaction(SIGUSR1, &sa, NULL);
#endif
    do {
        if (!cache_fill(s)) {
#if FORKED_CACHE
            // Let signal wake us up, we cannot leave this
            // enabled since we do not handle EINTR in most places.
            // This might need extra code to work on BSD.
            sa.sa_handler = dummy_sighandler;
            sigaction(SIGUSR1, &sa, NULL);
#endif
            if (sleep_count < INITIAL_FILL_USLEEP_COUNT) {
                sleep_count++;
                usec_sleep(INITIAL_FILL_USLEEP_TIME);
            } else
                usec_sleep(FILL_USLEEP_TIME); // idle
#if FORKED_CACHE
            sa.sa_handler = SIG_IGN;
            sigaction(SIGUSR1, &sa, NULL);
#endif
        } else
            sleep_count = 0;
    } while (cache_execute_control(s));
}

/**
 * \return 1 on success, 0 if the function was interrupted and -1 on error
 */
int stream_enable_cache(stream_t *stream,int size,int min,int seek_limit){
  int ss=(stream->type==STREAMTYPE_VCD)?VCD_SECTOR_DATA:STREAM_BUFFER_SIZE;
  int res = -1;
  cache_vars_t* s;

  DEBUG(kprintf("%s stream_enable_cache\n", CURRENT_TASK));

  if (stream->flags & STREAM_NON_CACHEABLE) {
    mp_msg(MSGT_CACHE,MSGL_STATUS,"\rThis stream is non-cacheable\n");
    return 1;
  }

  s=cache_init(size,ss);
  if(s == NULL) return -1;
  stream->cache_data=s;
  s->stream=stream; // callback
  s->seek_limit=seek_limit;


  //make sure that we won't wait from cache_fill
  //more data than it is allowed to fill
  if (s->seek_limit > s->buffer_size - s->fill_limit ){
	 s->seek_limit = s->buffer_size - s->fill_limit;
  }
  if (min > s->buffer_size - s->fill_limit) {
	 min = s->buffer_size - s->fill_limit;
  }
  // to make sure we wait for the cache process/thread to be active
  // before continuing
  if (min <= 0)
    min = 1;

#if FORKED_CACHE
  if((stream->cache_pid=fork())){
    if ((pid_t)stream->cache_pid == -1)
      stream->cache_pid = 0;
#else
  {
    stream_t* stream2=malloc(sizeof(stream_t));
    memcpy(stream2,s->stream,sizeof(stream_t));
    s->stream=stream2;
#if defined(__MINGW32__)
    stream->cache_pid = _beginthread( ThreadProc, 0, s );
#elif defined(__OS2__)
    stream->cache_pid = _beginthread( ThreadProc, NULL, 256 * 1024, s );
#elif defined(__MORPHOS__)
	DEBUG(kprintf("%s buffer_size = %d seek_limit = %d fill_limit = %d min = %d\n", CURRENT_TASK, s->buffer_size, s->seek_limit, s->fill_limit, min));

	if (!(stream->StartupMsg = (struct CustomMsg *) AllocVec( sizeof(struct CustomMsg), MEMF_PUBLIC ) ) )return -1;

	stream->StartupMsg->Msg.mn_Node.ln_Type = NT_MESSAGE;
	stream->StartupMsg->Msg.mn_ReplyPort = CreateMsgPort();
	stream->StartupMsg->Msg.mn_Length = sizeof(struct CustomMsg);
	stream->StartupMsg->s = (APTR) s;

	if (!stream->StartupMsg->Msg.mn_ReplyPort ) return -1;

	DEBUG(kprintf("%s About to create CacheTask\n", CURRENT_TASK));

	if(s->stream->type == STREAMTYPE_STREAM)
	{
		 // backup old socketbase that will be reopened in cache task
		 oldSocketBase = SocketBase;

		 // prepare to pass socket to cachetask
		 fdkey = ReleaseCopyOfSocket(stream->fd, time(NULL));

		 if(fdkey == -1)
		 {
			 return -1;
		 }

		 stream->fd = -1;
	}
	else if(s->stream->type == STREAMTYPE_DVD || s->stream->type == STREAMTYPE_DVDNAV)
	{
		 gdvdcss_close();
	}


   if ( ! (cachetask = CreateNewProcTags(  
#ifdef __amigaos4__
						NP_StackSize,              512*1024,
                        			NP_UserData,               (IPTR)stream->StartupMsg,
#else
                        			NP_CodeType,               CODETYPE_PPC,
						NP_StartupMsg,             (ULONG) stream->StartupMsg,
						NP_PPCStackSize,           512*1024,
#endif
						NP_Entry,                  (ULONG) CacheTask,
						//NP_Priority,               1,
						NP_Input,                  Input(),
						NP_CloseInput,             FALSE,
						NP_Output,                 Output(),
						NP_CloseOutput,            FALSE,
						NP_Name,                   "MPlayer Cache task",
						NP_Child,                  TRUE,
						TAG_DONE) ) ) return -1; // Unable to create the Process



	DEBUG(kprintf("%s CacheTask created\n", CURRENT_TASK));

	stream->cache_pid = 1;
#else
    {
    pthread_t tid;
    pthread_create(&tid, NULL, ThreadProc, s);
    stream->cache_pid = 1;
    }
#endif
#endif
    if (!stream->cache_pid) {
        mp_msg(MSGT_CACHE, MSGL_ERR,
               "Starting cache process/thread failed: %s.\n", strerror(errno));
        goto err_out;
    }
    // wait until cache is filled at least prefill_init %
    mp_msg(MSGT_CACHE,MSGL_V,"CACHE_PRE_INIT: %"PRId64" [%"PRId64"] %"PRId64"  pre:%d  eof:%d  \n",
	(int64_t)s->min_filepos,(int64_t)s->read_filepos,(int64_t)s->max_filepos,min,s->eof);
	while(s->read_filepos<s->min_filepos || s->max_filepos-s->read_filepos<min){
	mp_msg(MSGT_CACHE,MSGL_STATUS,MSGTR_CacheFill,
	    100.0*(float)(s->max_filepos-s->read_filepos)/(float)(s->buffer_size),
	    (int64_t)s->max_filepos-s->read_filepos
	);
	if(s->eof) break; // file is smaller than prefill size
	if(stream_check_interrupt(PREFILL_SLEEP_TIME)) {
	  res = 0;
	  goto err_out;
        }
    }
    mp_msg(MSGT_CACHE,MSGL_STATUS,"\n");
    return 1; // parent exits

  }

#if FORKED_CACHE
  signal(SIGTERM,exit_sighandler); // kill
  cache_mainloop(s);
  // make sure forked code never leaves this function
  exit(0);
#endif

err_out:
    cache_uninit(stream);
    return res;
}

#if !FORKED_CACHE
#if defined(__MINGW32__) || defined(__OS2__)
static void ThreadProc( void *s ){
  cache_mainloop(s);
  _endthread();
}
#else
static void *ThreadProc( void *s ){
  cache_mainloop(s);
  return NULL;
}
#endif
#endif

int cache_stream_fill_buffer(stream_t *s){
  int len;
  int sector_size;

  DEBUG(kprintf("%s cache_stream_fill_buffer s->type = %d eof = %d\n", CURRENT_TASK, s->type, s->eof));
  DEBUG(kprintf("%s buf_pos = %d buf_len = %d\n", CURRENT_TASK, s->buf_pos, s->buf_len));

  //if(s->eof){ s->buf_pos=s->buf_len=0; return 0; } // XXX: still needed?
  if(!s->cache_pid || FindTask(NULL) != maintask)
  {
	DEBUG(kprintf("%s stream_fill_buffer\n", CURRENT_TASK));
	return stream_fill_buffer(s);
  }

  DEBUG(kprintf("%s cache_stream_fill_buffer pos = %lld read_filepos %lld\n", CURRENT_TASK, (quad_t) s->pos, ((cache_vars_t*)s->cache_data)->read_filepos));

  if(s->pos!=((cache_vars_t*)s->cache_data)->read_filepos) mp_msg(MSGT_CACHE,MSGL_ERR,"!!! read_filepos differs!!! report this bug...\n");
  sector_size = ((cache_vars_t*)s->cache_data)->sector_size;
/*
  if (sector_size > STREAM_MAX_SECTOR_SIZE) {
    mp_msg(MSGT_CACHE, MSGL_ERR, "Sector size %i larger than maximum %i\n", sector_size, STREAM_MAX_SECTOR_SIZE);
    sector_size = STREAM_MAX_SECTOR_SIZE;
  }
*/
  len=cache_read(s->cache_data,s->buffer, sector_size);
  //printf("cache_stream_fill_buffer->read -> %d\n",len);

  if(len<=0){ s->eof=1; s->buf_pos=s->buf_len=0; return 0; }
  s->eof=0;
  s->buf_pos=0;
  s->buf_len=len;
  s->pos+=len;
//  printf("[%d]",len);fflush(stdout);
  if (s->capture_file)
    stream_capture_do(s);
  return len;

}

int cache_stream_seek_long(stream_t *stream,quad_t pos){
  cache_vars_t* s;
  quad_t newpos;

  DEBUG(kprintf("%s cache_stream_seek_long pos = %lld\n", CURRENT_TASK, pos));

  if(!stream->cache_pid || FindTask(NULL) != maintask)
  {
	  DEBUG(kprintf("%s stream_seek_long\n", CURRENT_TASK));
	  return stream_seek_long(stream,pos);
  }

  s=stream->cache_data;
//  s->seek_lock=1;

  mp_msg(MSGT_CACHE,MSGL_DBG2,"CACHE2_SEEK: 0x%"PRIX64" <= 0x%"PRIX64" (0x%"PRIX64") <= 0x%"PRIX64"  \n",s->min_filepos,pos,s->read_filepos,s->max_filepos);

  newpos=pos/s->sector_size; newpos*=s->sector_size; // align
  stream->pos=s->read_filepos=newpos;
  s->eof=0; // !!!!!!!
  cache_wakeup(stream);

  cache_stream_fill_buffer(stream);

  pos-=newpos;
  if(pos>=0 && pos<=stream->buf_len){
	stream->buf_pos=pos; // byte position in sector
	return 1;
  }

//  stream->buf_pos=stream->buf_len=0;
//  return 1;

  //mp_msg(MSGT_CACHE,MSGL_V,"cache_stream_seek: WARNING! Can't seek to 0x%"PRIX64" !\n",(int64_t)(pos+newpos));
  return 0;
}

int cache_do_control(stream_t *stream, int cmd, void *arg) {
  int sleep_count = 0;
  cache_vars_t* s = stream->cache_data;
  switch (cmd) {
    case STREAM_CTRL_SEEK_TO_TIME:
      s->control_double_arg = *(double *)arg;
      s->control = cmd;
      break;
    case STREAM_CTRL_SEEK_TO_CHAPTER:
    case STREAM_CTRL_SET_ANGLE:
      s->control_uint_arg = *(unsigned *)arg;
      s->control = cmd;
      break;
// the core might call these every frame, they are too slow for this...
// __MORPHOS__  it wasn't enabled, which fucks up dvd time display
	case STREAM_CTRL_GET_CURRENT_TIME:
	  *(double *)arg = s->stream_current_time;
	  return s->stream_current_time ? STREAM_OK : STREAM_UNSUPPORTED;
// __MORPHOS__
    case STREAM_CTRL_GET_TIME_LENGTH:
//	  case STREAM_CTRL_GET_CURRENT_TIME:
      *(double *)arg = s->stream_time_length;
      return s->stream_time_length ? STREAM_OK : STREAM_UNSUPPORTED;
    case STREAM_CTRL_GET_NUM_CHAPTERS:
    case STREAM_CTRL_GET_CURRENT_CHAPTER:
    case STREAM_CTRL_GET_ASPECT_RATIO:
    case STREAM_CTRL_GET_NUM_ANGLES:
    case STREAM_CTRL_GET_ANGLE:
    case -2:
      s->control = cmd;
      break;
    default:
      return STREAM_UNSUPPORTED;
  }
  cache_wakeup(stream);
  while (s->control != -1) {
    if (sleep_count++ == 1000)
      mp_msg(MSGT_CACHE, MSGL_WARN, "Cache not responding!\n");
    if (stream_check_interrupt(CONTROL_SLEEP_TIME)) {
      s->eof = 1;
      return STREAM_UNSUPPORTED;
    }
  }
  switch (cmd) {
    case STREAM_CTRL_GET_TIME_LENGTH:
    case STREAM_CTRL_GET_CURRENT_TIME:
    case STREAM_CTRL_GET_ASPECT_RATIO:
      *(double *)arg = s->control_double_arg;
      break;
    case STREAM_CTRL_GET_NUM_CHAPTERS:
    case STREAM_CTRL_GET_CURRENT_CHAPTER:
    case STREAM_CTRL_GET_NUM_ANGLES:
    case STREAM_CTRL_GET_ANGLE:
      *(unsigned *)arg = s->control_uint_arg;
      break;
    case STREAM_CTRL_SEEK_TO_CHAPTER:
    case STREAM_CTRL_SEEK_TO_TIME:
    case STREAM_CTRL_SET_ANGLE:
      if (s->control_res != STREAM_UNSUPPORTED)
          stream->pos = s->read_filepos = s->control_new_pos;
      break;
  }
  return s->control_res;
}

#endif
